/***************************************************************************
 *   Copyright (C) 2008 by Stephane List                                   *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QNetworkInterface> 
#include <QtGui>
#include <QSizePolicy>
#include <QDateTime>
#include <QBrush>
#include "plc.h"
#include "plconfig.h"

#include <ctype.h>

enum {
COLUMN_DEVICE_NAME = 0,
//COLUMN_PASSWORD,
//COLUMN_QUALITY,
COLUMN_RX_RATE,
COLUMN_TX_RATE,
COLUMN_MAC,
COLUMN_LAST_DATE,
COLUMN_LAST};

#define AUTOSCAN_TEMPO  2 // scan every x seconds
#define AUTOSCAN_REMOVE 5 // remove old entries after x seconds of inactivity
//#define CLOSE_BUTTON 

plc::plc()
{
	if_count = 0;
	netfd = -1;
	MyMac = "";
	autoscan = true;

	QCoreApplication::setOrganizationName("Easyvisio");
	QCoreApplication::setOrganizationDomain("easyvisio.com");
	QCoreApplication::setApplicationName("plc");
	settings = new QSettings();




	IfGroup = new QGroupBox(tr("Interfaces"));
	QVBoxLayout *IfLayout = new QVBoxLayout;

	foreach (QNetworkInterface interface, QNetworkInterface::allInterfaces()) {
		if ( !(interface.flags() & QNetworkInterface::IsLoopBack) &&
				(interface.flags() & QNetworkInterface::IsUp) &&
				(interface.flags() & QNetworkInterface::IsRunning))
		{
			if_count++;
			printf("Interface found: %s\n", interface.name().toAscii().constData());

			QCheckBox *tmp = new QCheckBox(interface.name());
			tmp->setChecked(true);
			IfLayout->addWidget(tmp);

			if (netfd == -1)
			{
				/* We will take the first IF, TODO take a list of all ifs */
				netfd = plc_init(interface.name());
			}
		}
	}
	if (if_count == 0)
	{
		QMessageBox::critical(this, tr("plc"), tr("No network interface found!\n"), QMessageBox::Ok);
		exit(0);
	}
	IfGroup->setLayout(IfLayout);
	if (netfd == -1)
	{
		QMessageBox::critical(this, tr("plc"), tr("Impossible to open socket, you need to be administrator!\n"), QMessageBox::Ok);
		exit(0);
	}
	setMinimumSize(800, 600);
	setWindowTitle(tr("plc"));
	createTab();

	printf("Chargement des anciens Devices vus dans plc\n");
	int i=0;
	MainTab_->ScannedTable->setSortingEnabled(false);
	foreach (QString mac, settings->childGroups()) {
		printf("Chargement des anciens Devices vus dans plc : %s (%d)\n", mac.toAscii().constData(), i);
		MainTab_->ScannedTable->setRowCount(i+1);
		settings->beginGroup(mac);
		MainTab_->ScannedTable->setItem(i, COLUMN_DEVICE_NAME,new QTableWidgetItem(settings->value("dev_name"," ").toString()));
		MainTab_->ScannedTable->setItem(i, COLUMN_MAC,new QTableWidgetItem(mac));
		MainTab_->ScannedTable->setItem(i, COLUMN_LAST_DATE,new QTableWidgetItem(settings->value("last_date"," ").toString()));
		MainTab_->ScannedTable->item(i, COLUMN_MAC)->setFlags(MainTab_->ScannedTable->item(i, COLUMN_MAC)->flags() & ~Qt::ItemIsEditable);
		MainTab_->ScannedTable->item(i, COLUMN_LAST_DATE)->setFlags(MainTab_->ScannedTable->item(i, COLUMN_LAST_DATE)->flags() & ~Qt::ItemIsEditable);
		MainTab_->ScannedTable->resizeColumnToContents(COLUMN_MAC);
		i++;
		settings->endGroup();
	}
	MainTab_->ScannedTable->setSortingEnabled(true);
	printf("Chargement des anciens Devices vus dans plc (fini)\n");

	listenerThread = new ListenerThread(this);

	connect(listenerThread, SIGNAL(Mac_detected(const QString &)), MainTab_, SLOT(add_mac(const QString &)));
	connect(listenerThread, SIGNAL(My_Mac_detected(const QString &, const QString &)), MainTab_, SLOT(add_my_mac(const QString &, const QString &)));
	connect(listenerThread, SIGNAL(Rate_detected(const QString &, int, int)), MainTab_, SLOT(add_rate(const QString &, int, int)));
	connect(listenerThread, SIGNAL(Date_detected(const QString &)), MainTab_, SLOT(add_date(const QString &)));
	connect(listenerThread, SIGNAL(Nek_changed_confirm()), PrivacyTab_, SLOT(display_nek_confirm()));
	listenerThread->start(QThread::LowPriority);

	scanThread = new ScanThread(this);
	connect(scanThread, SIGNAL(remove_old()), MainTab_, SLOT(remove_old()));
	scanThread->start(QThread::LowPriority);
}

void plc::createTab()
{
  tabWidget = new QTabWidget(this);
  MainTab_ = new MainTab(this);
  PrivacyTab_ = new PrivacyTab(this);

  tabWidget->addTab(MainTab_, tr("Main"));
  tabWidget->addTab(PrivacyTab_, tr("Privacy"));
//  tabWidget->addTab(new DiagnosticsTab(), tr("Diagnostics"));
  tabWidget->addTab(new AboutTab(this), tr("About"));

#ifdef CLOSE_BUTTON
  buttonBox = new QDialogButtonBox(QDialogButtonBox::Close
      | QDialogButtonBox::Cancel);

  connect(buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
  connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));
#endif
  QVBoxLayout *mainLayout = new QVBoxLayout(this);
  mainLayout->addWidget(tabWidget);
  if (if_count>1)
    mainLayout->addWidget(IfGroup);
#ifdef CLOSE_BUTTON
  mainLayout->addWidget(buttonBox);
#endif
  setLayout(mainLayout);
}

void plc::setMyMac(const QString& mac)
{
  MyMac = mac;
}

QString& plc::getMyMac()
{
  return MyMac;
}

int plc::getfd()
{
  return netfd;
}

plc::~plc()
{
  scanThread->quit();
  listenerThread->quit();
  plc_deinit();
}

MainTab::MainTab(plc *parent) : QTabWidget(parent)
{
  int h;

  Parent = parent;

  QLabel *Label1 = new QLabel(tr("Local Device on your computer:"));

  ConnectedTable = new QTableWidget(1,2);
  QStringList htab;
  htab << "MAC Address" << "Firmware";
  ConnectedTable->setHorizontalHeaderLabels(htab);
#if QT_VERSION >= 0x040300
  ConnectedTable->setCornerButtonEnabled(false);
#endif
  ConnectedTable->resizeColumnsToContents();
  ConnectedTable->setRowCount(1);

  h = ConnectedTable->rowHeight(0);
  h += ConnectedTable->horizontalHeader()->height();
  ConnectedTable->setMaximumHeight(h);
  ConnectedTable->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
  ConnectedTable->horizontalHeader()->setStretchLastSection(true);
  ConnectedTable->setEditTriggers(QAbstractItemView::NoEditTriggers);

  QLabel *Label2 = new QLabel(tr("Device(s) detected:"));

  ScannedTable = new QTableWidget();
  ScannedTable->setMinimumSize(0, 400);
  ScannedTable->setRowCount(0);
  ScannedTable->setColumnCount(COLUMN_LAST);
//  ScannedTable->setEditTriggers(QAbstractItemView::NoEditTriggers);
  QStringList htab2;
  htab2 << "Device Name" 
//	<< "Password"
//	<< "Quality"
	<< "RX Rate(Mbps)"
	<< "TX Rate(Mbps)"
	<< "MAC Address"
	<< "Last Date";
  ScannedTable->setHorizontalHeaderLabels(htab2);
  ScannedTable->horizontalHeader()->setStretchLastSection(true);

  ScannedTable->sortItems(0, Qt::AscendingOrder);
  ScannedTable->setSortingEnabled(true);
#if QT_VERSION >= 0x040300
  ScannedTable->setCornerButtonEnabled(false);
#endif
// TODO
// ScannedTable->hideColumn(COLUMN_LAST_DATE);
  ScannedTable->resizeColumnsToContents();

  QVBoxLayout *mainLayout = new QVBoxLayout;
  mainLayout->addWidget(Label1);
  mainLayout->addWidget(ConnectedTable);
  mainLayout->addWidget(Label2);
  mainLayout->addWidget(ScannedTable);

#if 0
  QTableWidget *TestTable = new QTableWidget();
  TestTable->setMinimumSize(0, 400);
  TestTable->setRowCount(5);
  TestTable->setColumnCount(COLUMN_LAST);
  mainLayout->addWidget(TestTable);
#endif

  mainLayout->addStretch(1);
  setLayout(mainLayout);
}

void MainTab::add_mac(const QString& mac)
{
  QString& my = Parent->getMyMac();

  QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);
  if (MyQList.size() == 0)
  {
    if ((! my.isEmpty()) && (my != mac))
    {
      int i = ScannedTable->rowCount();
      int j;
      ScannedTable->setSortingEnabled(false);
      ScannedTable->setRowCount(i+1);
      ScannedTable->setItem(i, COLUMN_DEVICE_NAME,new QTableWidgetItem(QString("Device %1").arg(i+1)));
      ScannedTable->setItem(i, COLUMN_MAC,new QTableWidgetItem(mac));
      QDateTime now = QDateTime::currentDateTime();
      ScannedTable->setItem(i, COLUMN_LAST_DATE,new QTableWidgetItem(now.toString("dddd d/M/yyyy hh:mm:ss")));
      for (j=0; j< COLUMN_LAST; j++)
      {
	      if (j != COLUMN_DEVICE_NAME)
	      {
		      ScannedTable->item(i, j)->setFlags(ScannedTable->item(i, j)->flags() & ~Qt::ItemIsEditable);
	      }
      }
      ScannedTable->resizeColumnToContents(COLUMN_MAC);
      ScannedTable->setSortingEnabled(true);
    }
    else
    {
      if (my.isEmpty())
      {
	plc_req_version(mac);
      }
    }
  }
}

void MainTab::add_my_mac(const QString &mac, const QString &firmware)
{
  QList<QTableWidgetItem*>MyQList = ConnectedTable->findItems(mac, Qt::MatchFixedString);

  if (MyQList.size() == 0)
  {
    int i = ConnectedTable->rowCount();
    ConnectedTable->insertRow(i);
    ConnectedTable->setItem(0,0,new QTableWidgetItem(mac));
    ConnectedTable->setItem(0,1,new QTableWidgetItem(firmware));
    ConnectedTable->resizeColumnToContents(0);
    Parent->setMyMac(mac);
  }
}

/*
 * mac = MAC address
 * rate = rate in Mbit/s
 * direction = 0 for RX, 1 for TX */
void MainTab::add_rate(const QString &mac, int rate, int direction)
{
	QString my = Parent->getMyMac();
	int col, i;
	QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);

	if (( my.isEmpty()) || (my == mac))
		return;

	if (direction == 0)
		col = COLUMN_RX_RATE;
	else
		col = COLUMN_TX_RATE;

	if (MyQList.size() == 0)
	{
		i = ScannedTable->rowCount();
		ScannedTable->setSortingEnabled(false);
		ScannedTable->setRowCount(i+1);

		QTableWidgetItem *item0 = new QTableWidgetItem(QString("Device %1").arg(i+1));
		ScannedTable->setItem(i, COLUMN_DEVICE_NAME, item0);

		ScannedTable->setItem(i, COLUMN_MAC,new QTableWidgetItem(mac));
		ScannedTable->item(i,COLUMN_MAC)->setFlags(ScannedTable->item(i, COLUMN_MAC)->flags() & ~Qt::ItemIsEditable);

		ScannedTable->setItem(i,col,new QTableWidgetItem(QString("%1").arg(rate)));
		ScannedTable->item(i,col)->setFlags(ScannedTable->item(i, col)->flags() & ~Qt::ItemIsEditable);

		(ScannedTable->item(i,col))->setTextAlignment(Qt::AlignRight);
		ScannedTable->resizeColumnToContents(COLUMN_DEVICE_NAME);
		ScannedTable->resizeColumnToContents(col);
		ScannedTable->resizeColumnToContents(COLUMN_MAC);
	}
	else /* We need to update the entry */
	{
		i = MyQList[0]->row();

		ScannedTable->setItem(i,col,new QTableWidgetItem(QString("%1").arg(rate)));
		(ScannedTable->item(i,col))->setTextAlignment(Qt::AlignRight);
		ScannedTable->item(i,col)->setFlags(ScannedTable->item(i, col)->flags() & ~Qt::ItemIsEditable);
		ScannedTable->resizeColumnToContents(col);
	}
}

// Update Last seen value
void MainTab::add_date(const QString &mac)
{
	int row;
	QSettings *settings = Parent->settings;

	QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);
	if (MyQList.size() != 0)
	{
		QDateTime now = QDateTime::currentDateTime();
		row = MyQList[0]->row();
		ScannedTable->setItem(row, COLUMN_LAST_DATE,new QTableWidgetItem(now.toString("dddd d/M/yyyy hh:mm:ss")));
		ScannedTable->item(row, COLUMN_LAST_DATE)->setFlags(ScannedTable->item(row, COLUMN_LAST_DATE)->flags() & ~Qt::ItemIsEditable);

		settings->beginGroup(ScannedTable->item(row, COLUMN_MAC)->text());
		settings->setValue("dev_name" , ScannedTable->item(row, COLUMN_DEVICE_NAME)->text());
		settings->setValue("last_date", ScannedTable->item(row, COLUMN_LAST_DATE)->text());
		settings->endGroup();
		settings->sync();
	}
}

void MainTab::remove_old()
{
	int row;

	QDateTime now = QDateTime::currentDateTime();

	for (row=0; row < (ScannedTable->rowCount()); row++)
	{
		if ((ScannedTable->item(row, COLUMN_LAST_DATE) != NULL) && (ScannedTable->item(row, COLUMN_LAST_DATE)->text().isEmpty() != true))
		{
			QDateTime last = QDateTime::fromString(ScannedTable->item(row, COLUMN_LAST_DATE)->text(), "dddd d/M/yyyy hh:mm:ss");
			if (last.isValid() && (last.secsTo(now) > AUTOSCAN_REMOVE))
			{
				ScannedTable->item(row, COLUMN_LAST_DATE)->setBackgroundColor(Qt::red);
				printf("remove_old (%d seconds)!!!\n", last.secsTo(now));
#if 0
				QString *nows = new QString(now.toString("dddd d/M/yyyy hh:mm:ss"));
				printf("now : %s\n", nows->toAscii().constData());

				QString *lasts = new QString(last.toString("dddd d/M/yyyy hh:mm:ss"));
				printf("last : %s\n", lasts->toAscii().constData());
#endif
			}
		}
	}
}

PrivacyTab::PrivacyTab(plc *parent) : QTabWidget(parent)
{
  Parent = parent;
  Nek_state = NEK_IDLE;
  QLabel *Label1 = new QLabel(tr("Use this screen to create a Private Network that provides you with extra security.\n"));

  QGroupBox *Group1 = new QGroupBox(tr("Private Network Name"));
  nekLineEdit = new QLineEdit;
  nekLineEdit->setText(QString("HomePlug"));
  nekLineEdit->setFocus();
  /*nekLineEdit->setGeometry(QRect(0, 0, 350, 25));
  QSizePolicy sizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
  nekLineEdit->setSizePolicy(sizePolicy);*/
  QLabel *orLabel = new QLabel(tr("OR"));
  QPushButton *pb1 = new QPushButton(tr("Use Default (Public Network)"), this);
  connect(pb1, SIGNAL(clicked()), this, SLOT(setDefaultNek()));

  QLabel *warnLabel = new QLabel(tr("Do not share the network name with others you do not want to be part of this network."));
  QGridLayout *nekLayout = new QGridLayout;
  nekLayout->addWidget(nekLineEdit, 0, 0);
  nekLayout->addWidget(orLabel, 0, 1);
  nekLayout->addWidget(pb1, 0, 2);
  nekLayout->addItem(new QSpacerItem(1,1,QSizePolicy::Expanding,QSizePolicy::Minimum), 0, 3);
  nekLayout->addWidget(warnLabel, 1, 0, 1, 4, 0);
  Group1->setLayout(nekLayout);

  QLabel *Label2 = new QLabel(tr("\nAfter setting the name above, choose how it will be applied below:\n"));

  QGroupBox *Group2 = new QGroupBox(tr("Place On This Network"));
  
  QLabel *Label3 = new QLabel(tr("Only the device attached to this computer (the Local Device),\n(Isolates this computer from others)"));
  Label3->setAlignment(Qt::AlignRight|Qt::AlignVCenter);
  
  QLabel *Label4 = new QLabel(tr("ALL devices whose Password has been entered.\n(They communicate together but are isolated from devices with a different network name)" ));
  Label4->setAlignment(Qt::AlignRight|Qt::AlignVCenter);
  Label4->setWordWrap(true);
  
  QPushButton *pb2 = new QPushButton(tr("Set Local Device Only"), this);
  connect(pb2, SIGNAL(clicked()), this, SLOT(setLocalDevice()));
  
  QLabel *Label5 = new QLabel(tr("OR" ));
  Label5->setAlignment(Qt::AlignHCenter|Qt::AlignVCenter);
  
  QPushButton *pb3 = new QPushButton(tr("Set All Devices (buggy, so disabled :-< )"), this);
  pb3->setEnabled(false);
  connect(pb3, SIGNAL(clicked()), this, SLOT(setAllDevices()));

  QGridLayout *placeLayout = new QGridLayout;
  placeLayout->addWidget(Label3, 0, 0);
  placeLayout->addWidget(Label4, 2, 0);
  placeLayout->addWidget(pb2,0, 1);
  placeLayout->addWidget(Label5,1, 1);
  placeLayout->addWidget(pb3,2, 1);
  Group2->setLayout(placeLayout);

  QVBoxLayout *mainLayout = new QVBoxLayout;
  mainLayout->addWidget(Label1);
  mainLayout->addWidget(Group1);
  mainLayout->addWidget(Label2);
  mainLayout->addWidget(Group2);
  mainLayout->addStretch(1);
  setLayout(mainLayout);
}

void PrivacyTab::setDefaultNek()
{
  nekLineEdit->setText(QString("HomePlug"));
}
    	
void PrivacyTab::setLocalDevice()
{
  QString& my = Parent->getMyMac();

  if (my.isEmpty())
  {
    QMessageBox::critical(this, tr("plc"), tr("No Local Device found!\n"), QMessageBox::Ok);
  }
  else
  {
    Nek_state = NEK_SENT_LOCAL;
    plc_set_network_key(my, nekLineEdit->text());
  }
}

void PrivacyTab::setAllDevices()
{
  QString& my = Parent->getMyMac();

  if (my.isEmpty())
  {
    QMessageBox::critical(this, tr("plc"), tr("No Local Device found!\n"), QMessageBox::Ok);
  }
  else
  {
    Nek_state = NEK_SENT_FIRST;
    plc_set_network_key(my, nekLineEdit->text());
  }
}

void PrivacyTab::display_nek_confirm()
{
	switch(Nek_state) {
		case NEK_SENT_LOCAL:
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully!\n"), QMessageBox::Ok);
			Nek_state = NEK_IDLE;
			break;
		case NEK_SENT_FIRST:
			Nek_state = NEK_SENT_SECOND;
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully!\nChange for all devices now."), QMessageBox::Ok);
			plc_broadcast_key();
			break;
		case NEK_SENT_SECOND:
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully on all devices!\n"), QMessageBox::Ok);
			Nek_state = NEK_IDLE;
			break;
		default:
			QMessageBox::warning(this, tr("plc"), tr("Private Network name changed by somebody in the network!\n"), QMessageBox::Ok);
			break;
	}
}

DiagnosticsTab::DiagnosticsTab(QWidget *parent) : QWidget(parent)
{
  QLabel *Label1 = new QLabel(tr("Remote Device History(Log)"));
  QVBoxLayout *mainLayout = new QVBoxLayout;
  mainLayout->addWidget(Label1);
  mainLayout->addStretch(1);
  setLayout(mainLayout);
}

AboutTab::AboutTab(plc *parent) : QTabWidget(parent)
{
  Parent = parent;

  QLabel *Label1 = new QLabel(tr("plc Version 0.0.9"));
  QLabel *Label2 = new QLabel(tr("Released under GNU GPL v3."));

  QGroupBox *pref = new QGroupBox(tr("Preferences"));

  QCheckBox *AutoScan;
  AutoScan = new QCheckBox(tr("Auto Scan"));
  AutoScan->setChecked(true);

  connect(AutoScan, SIGNAL(toggled(bool)), this, SLOT(AutoScan_toggled(bool)));

  QVBoxLayout *prefLayout = new QVBoxLayout;
  prefLayout->addWidget(AutoScan);
  pref->setLayout(prefLayout);

  QVBoxLayout *mainLayout = new QVBoxLayout;
  mainLayout->addWidget(Label1);
  mainLayout->addWidget(Label2);
  mainLayout->addWidget(pref);
  mainLayout->addStretch(1);
  setLayout(mainLayout);
}

void AboutTab::AutoScan_toggled(bool val)
{
  Parent->autoscan = val;
}

ScanThread::ScanThread(plc *parent)
:QThread(parent)
{
}

void ScanThread::run()
{
  plc *p = qobject_cast<plc *>(parent());

  forever
  {
    if (p->autoscan)
    {
      if ((p->getMyMac()).isEmpty())
      {
	plc_send_mm(); // find the adapter on witch we're attached
      }
      else
      {
	plc_get_channel_capacities(); // find connected (through Homeplug) devices and data rate
      }
    }
    else
    {
      printf("Autoscan off\n");
    }
    sleep(AUTOSCAN_TEMPO);

    emit remove_old();
    //	plc_req_param_stat();
    //	plc_req_intellon_specific();
    //	plc_req_perf_stat();
    //	plc_req_channel_estimation();
  }
}

ListenerThread::ListenerThread(plc *parent)
:QThread(parent)
{
}

int ListenerThread::receive(int fd, unsigned char* buf)
{
  ssize_t rdlen;
  struct sockaddr_ll addr;
  socklen_t addrlen = sizeof(addr);

  rdlen = recvfrom(fd, buf+ETHER_HDR_LEN, buflen-ETHER_HDR_LEN, MSG_TRUNC, (struct sockaddr *)&addr, &addrlen);
  if (rdlen < 0)
  {
    return -1;
  }
  if (rdlen > (ssize_t)(buflen-ETHER_HDR_LEN))
  {
    fprintf(stderr, "received jumbo frame of %u bytes len, truncated\n", (unsigned int)rdlen);
  }
  memcpy(buf+6, &addr.sll_addr, 6);
  if (addr.sll_pkttype == PACKET_OUTGOING)
  {
    printf("Ignore outgoing packet\n");
    return -1;
  }
  if (addr.sll_protocol != htons(ETHERTYPE_INTELLON))
  {
    printf("Ignore non Intellon packet\n");
    return -1;
  }
  return 0;
}

void ListenerThread::run()
{
  unsigned char *framebuf;
  char macbuf[20];
  int netfd;
  int i;
  unsigned int j;
  unsigned char mid;
  unsigned char tmp;
  unsigned char tmp2;

  unsigned char melen, metype;
  unsigned char *macmgmt;

  framebuf = (unsigned char*)malloc((size_t)buflen);

  plc *p = qobject_cast<plc *>(parent());
  netfd = p->getfd();

  while(1)
  {
    if (receive(netfd, framebuf) >=0)
    {
      /* It's an intellon packet - read MAC management entries */
      printf("From %s : ", format_mac_addr(&framebuf[6], macbuf));

      QString NewMac(format_mac_addr(&framebuf[6], macbuf));

      emit Mac_detected(NewMac);

      melen = framebuf[16];
      metype = framebuf[15] & 0x1F; /* 5 bits */
      printf("0x%X ", metype);
      switch (metype)
      {
	case 0x00:
	  printf("Request Channel Estimation\n");
	  break;
	case 0x01:
	  printf("Channel Estimation response\n");
	  break;
	case 0x02:
	  printf("Vendor Specific Parameters\n\t");
	  for (i = 0; i< 15; i++)
	  {
	    printf("0x%X ", framebuf[15+i]);
	  }
	  printf("\n");
	  mid = framebuf[20] & 0x7F; // remove first bit (1 : response)
	  printf("\tMID = %d", mid);
	  if (!(framebuf[20] & 0x80))
	  {
	    printf("(request)\n");
	  }
	  else
	  {
		emit Date_detected(NewMac);
		printf("(response)\n");
	  }
	  switch (mid)
	  {
	    if ((framebuf[17] != 0x00) | (framebuf[18] != 0x04) | (framebuf[19] != 0x87))
	    {
	      printf("Warning, Vendor != Intellon\n");
	    }


	    case 16:
	    printf("\tProtocol Version Flag = %d\n", framebuf[20] & 0x02);
	    if ((framebuf[20] & 0x02) == 0)
	    {
	      printf("\tMAC Version 1.2\n");
	    }
	    else
	    {
	      printf("\tMAC Version TODO!\n"); // TODO
	    }
	    if ((framebuf[20] & 0x01) == 0)
	    {
	      printf("\tConnection : local\n");
	    }
	    else
	    {
	      printf("\tConnection : remote\n");
	    }
	    tmp = framebuf[25];
	    printf("\tManufacturerStringLen = %d\n", tmp);
	    printf("\tManufacturer = ");
	    for (i=0; i<tmp; i++)
	    {
	      printf("%c", framebuf[25+1+i]);
	    }
	    printf("\n");
	    tmp2 = framebuf[25 + tmp +1];
	    //	tmp2 = 7; /* seems buggy, so force to 7 */
	    printf("\tProductNameStringLen = %d\n", tmp2);
	    printf("\tProductName = ");
	    for (i=0; i<tmp2; i++)
	    {
	      if ( isprint(framebuf[25+ 2 + tmp + i]))
	      {

		printf("%c", framebuf[25+ 2 + tmp + i]);
	      }
	      else 
	      {
		//		tmp2 = 0; // stop loop because of invalid character
	      }
	    }
	    printf("\n");
	    break;
	    case 24:
	      if ( NewMac != p->getMyMac())
		      break; // don't trace RX TX values of all devices

	    switch (capabilities_flag) {
	      unsigned char n;

	      case CAPABILITIES_RX:
	      printf("\treceived channel capacities response mme for RX\n");

	      if ((melen-4)%14 != 0)
	      {
		printf("Error, ignore frame!\n");
		break;
	      }
	      n = (melen -4) / 14;
	      printf("\t%d RX station(s)\n", n);
	      for (i=0; i<n; i++)
	      {
		printf("\tStation %d\n", i);
		printf("\t\tMAC %02x:%02x:%02x:%02x:%02x:%02x\n",
		    framebuf[21+i*14],
		    framebuf[22+i*14],
		    framebuf[23+i*14],
		    framebuf[24+i*14],
		    framebuf[25+i*14],
		    framebuf[26+i*14]);
		j = ex_word(&framebuf[27+i*14]);
		//printf("\t\tRxBytesPerBlock = %d\n", j);
		printf("\t\tRxBytesRate = %d MBytes/s\n", j/336);
		printf("\t\tRxBitRate   = %d MBits/s\n", 8*j/336);
		//printf("\t\tRxClockError = %d\n", *(unsigned int*)(&framebuf[29+i*14]));
		// ignore the 2 following bytes (reserved)

//		if ( NewMac == p->getMyMac())
		{
		  QString Mac(format_mac_addr(&framebuf[21+i*14], macbuf));
		  emit Rate_detected(Mac, 8*j/336, 0);
		}
	      }
	      break;
	      case CAPABILITIES_TX:
	      printf("\treceived channel capacities response mme for TX\n");
	      if ((melen-4)%8 != 0)
	      {
		printf("Error, ignore frame !\n");
		break;
	      }
	      n = (melen -4) / 8;
	      printf("\t%d TX station(s)\n", n);
	      for (i=0; i<n; i++)
	      {
		printf("\tStation %d\n", i);
		printf("\t\tMAC %02x:%02x:%02x:%02x:%02x:%02x\n",
		    framebuf[21+i*8],
		    framebuf[22+i*8],
		    framebuf[23+i*8],
		    framebuf[24+i*8],
		    framebuf[25+i*8],
		    framebuf[26+i*8]);
		j = ex_word(&framebuf[27+i*8]);
		printf("\t\tTxBytesPerBlock = %d\n", j);
		printf("\t\tTxBytesRate = %d MBytes/s\n", j/336);
		printf("\t\tTxBitRate   = %d MBits/s\n", 8*j/336);
		if ( NewMac == p->getMyMac())
		{
		  QString Mac(format_mac_addr(&framebuf[21+i*8], macbuf));
		  emit Rate_detected(Mac, 8*j/336, 1);
		}
	      }
	      break;
	      default:
	      printf("\tError\n");
	      break;
	    }
	    break;
	    default:
	    printf("received unknow mid !\n");
	    break;
	  }
	  break;
	case 0x03:
	  printf("Replace Bridge Address\n");
	  break;
	case 0x04:
	  printf("Set Network Encryption Key\n");
	  break;
	case 0x05:
	  printf("Multicast with Response\n");
	  break;
	case 0x06:
	  printf("Confirm Network Encryption Key\n");
	  emit Nek_changed_confirm();
	  break;
	case 0x07:
	  printf("Parameters and Statistics request\n");
	  break;
	case 0x08:
	  printf("Parameters and Statistics response\n");
	  printf("\tlen = %d\n", framebuf[16]);
	  macmgmt=&framebuf[15];
	  printf(		"\tTx ACK Counter:             %u\n"
	      "\tTx NACK Counter:            %u\n"
	      "\tTx FAIL Counter:            %u\n"
	      "\tTx Contention Loss Counter: %u\n"
	      "\tTx Collision Counter:       %u\n"
	      "\tTx CA3 Latency Counter:     %u\n"
	      "\tTx CA2 Latency Counter:     %u\n"
	      "\tTx CA1 Latency Counter:     %u\n"
	      "\tTx CA0 Latency Counter:     %u\n"
	      "\tRx Cumul. Bytes per 40-symbol Packet Counter: %lu\n",
	      ex_word(&macmgmt[2]), ex_word(&macmgmt[4]), ex_word(&macmgmt[6]),
	      ex_word(&macmgmt[8]), ex_word(&macmgmt[10]), ex_word(&macmgmt[12]),
	      ex_word(&macmgmt[14]), ex_word(&macmgmt[16]), ex_word(&macmgmt[18]),
	      ex_long(&macmgmt[20]));
	  break;
	case 0x13:
	  printf("Get Device Version response\n");
	  if (framebuf[16] == 0x24)
	  {
	    printf("\tFirmware = %s\n", &framebuf[17]);
	    QString firmware((char *)&framebuf[17]);
	    emit My_Mac_detected(NewMac, firmware);
	  }
	  break;
	case 0x1a:		/* Intellon specific network statistics */
	  if (!(framebuf[17] & 0x80)) 		/* Really a response? */
	  {
	    printf("YYYYYYYYYYYYYYYEEEEEEEEEEEEEESSSSSSSSSSSSSSS ! Intellon-specific network statistics\n");
	  }
	  else
	  {
	    printf("Intellon-specific network statistics request\n");
	  }
	  break;
	case 0x1c:
	  printf("hip hip hip hourra !!!! \n");
	  break;
	case 0x1f:
	  printf("Set transmit characteristics\n");
	  break;
	default:
	  printf("Unknown response (METYPE = 0x%02x)\n", metype);
      }
    }
  }
}
