/***************************************************************************
 *   Copyright (C) 2007 by Stephane List   *
 *   slist@lilotux.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef PLC_H
#define PLC_H

#include <QMainWindow>
#include <QCloseEvent>
#include <QLineEdit>
#include <QTabWidget>
#include <QTableWidget>
#include <QDialogButtonBox>
#include <QDialog>
#include <QGroupBox>
#include <QThread>
#include <QString>
#include <QSettings>

//TODO : Use this instead of trying eth0,eth1...
//#include <QNetworkInterface>

//TODO : remove Linux specific !
//htons
#include <arpa/inet.h>

//sockaddr_ll
#include <netpacket/packet.h> 

//ETHER_HDR_LEN
#include <net/ethernet.h>

class QAction;
class QMenu;
class QTextEdit;

class MainTab;
class PrivacyTab;
class AboutTab;

enum {
  NEK_IDLE = 0,
  NEK_SENT_LOCAL,
  NEK_SENT_FIRST,
  NEK_SENT_SECOND
  };

class plc:public QDialog
{
  Q_OBJECT

  public:
    plc();
    ~plc();
    int getfd();
    QString& getMyMac();
    void setMyMac(const QString& mac);
    bool autoscan;
    QSettings *settings;

    // TODO : add something for a clean deinit...

  private:
    void createTab();
    QTabWidget *tabWidget;
    QDialogButtonBox *buttonBox;
    QThread *scanThread;
    QThread *listenerThread;
    int netfd;
    MainTab *MainTab_;
    PrivacyTab *PrivacyTab_;
    QGroupBox *IfGroup;
    int if_count;
    QString MyMac;
};

class MainTab : public QTabWidget
{
	Q_OBJECT

	public:
		MainTab(plc *parent);
		plc* Parent;
		QTableWidget *ConnectedTable;
		QTableWidget *ScannedTable;

	public slots:
		void add_mac(const QString & mac);
		void add_my_mac(const QString & mac, const QString & firmware);
		void add_rate(const QString & mac, int rate, int dir);
		void add_date(const QString & mac);
		void remove_old();
};


class PrivacyTab : public QTabWidget
{
  Q_OBJECT
  public:
    PrivacyTab(plc *parent);
    plc* Parent;

  private:
    QLineEdit *nekLineEdit;
    char Nek_state;

  public slots:
    void setDefaultNek();
    void setLocalDevice();
    void setAllDevices();
    void display_nek_confirm();
};

class DiagnosticsTab : public QWidget
{
	Q_OBJECT

	public:
		DiagnosticsTab(QWidget *parent = 0);
};

class AboutTab : public QTabWidget
{
	Q_OBJECT

	public:
		AboutTab(plc *parent);
		plc* Parent;

	public slots:
		void AutoScan_toggled(bool val);

};

class ScanThread : public QThread
{
	Q_OBJECT
	
	public:
		ScanThread(plc *parent = 0);
		void run();
	signals:
		void remove_old();
};

class ListenerThread : public QThread
{
	Q_OBJECT

	public:
		ListenerThread(plc *parent = 0);
		void run();
		int receive(int, unsigned char*);
	signals:
		void Mac_detected(const QString & mac);
		void Date_detected(const QString & mac);
		void Rate_detected(const QString & mac, int rate, int dir);
		void My_Mac_detected(const QString & mac, const QString & firmware);
		void Nek_changed_confirm();
};

#endif
